/**
 * Benchmark Test Source File
 *
 * @author: W. David Frohlingsdorf
 * @date: 2016
 * @purpose: Dissertation, University of Glasgow
 * 
 * This source file contains a micro-benchmark for
 * locking protocols. The thread library and the 
 * locking protocols API have to be defined in 
 * benchmark.h and necessary libraries have to 
 * be linked to during compilation.
 * 
 * There are currently 4 tests in this benchmark:
 * 1. A single thread locks and unlocks (in stack 
 *    fashion) TEST1_N_LOCKS. Measured time.
 * 2. A single thread locks and unlocks a single
 *    lock TEST2_N_RUNS times. Measured time.
 * 3. Deadlock two threads. Wait for TEST3_TIMEOUT*
 *    TEST_MAX_ATTEMPTS to timeout. Record result.
 * 4. Use three threads to simulate starvation.
 *    Wait for TEST4_MAX_ATTEMPTS*TEST4_TIMEOUT
 *    to timeout. Record result.
 * */

#include <iostream>
#include <ctime>
#include <cstdlib>
#include <unistd.h>
using namespace std;

#include "benchmark.h"


// * * Constants * * //
#define BENCHMARK_VERSION    "1.0"
// * * Other benchmark constants * * //
#define TEST1_N_LOCKS        100000
#define TEST2_N_RUNS         100000 
#define TEST3_MAX_ATTEMPTS   5
#define TEST3_TIMEOUT        2
#define TEST4_MAX_ATTEMPTS   10
#define TEST4_TIMEOUT        2

// * * Required static variables * * //
static BENCHMARK_LOCK_TYPE locks1[TEST1_N_LOCKS];
static BENCHMARK_LOCK_TYPE lock2;
static BENCHMARK_LOCK_TYPE lock3A, lock3B;
static BENCHMARK_LOCK_TYPE lock4A, lock4B;
static long startTime;
static volatile char testFinished;
static volatile char turn;

// * * Prototype-Declarations * * //
static int test1();
static int test2();
static int test3();
static int test4();
static FUNC_RET_TYPE run1(FUNC_ARG_TYPE);
static FUNC_RET_TYPE run2(FUNC_ARG_TYPE);
static FUNC_RET_TYPE run3A(FUNC_ARG_TYPE);
static FUNC_RET_TYPE run3B(FUNC_ARG_TYPE);
static FUNC_RET_TYPE run4A(FUNC_ARG_TYPE);
static FUNC_RET_TYPE run4B(FUNC_ARG_TYPE);
static FUNC_RET_TYPE run4C(FUNC_ARG_TYPE);
static void startClock();
static void stopClock(int testN);

// * * Main only triggers tests and stops execution if a test fails * * //
int main()
{
  cout << "Locking Protocol Benchmark Test - Version: " << BENCHMARK_VERSION << endl;

  if(!test1())
  {
    cout << "Test 1 FAILED! Aborting execution... " << endl;
    exit(0);
  }
  
  if(!test2())
  {
    cout << "Test 2 FAILED! Aborting execution... " << endl;
    exit(0);
  }

  if(!test3())
  {
    cout << "Test 3 FAILED! Aborting execution... " << endl;
    exit(0);
  }
  
  if(!test4())
  {
    cout << "Test 4 FAILED! Aborting execution... " << endl;
    exit(0);
  }

}

int test1()
{
  BENCHMARK_THREAD_TYPE t; //Single thread
  BENCHMARK_INIT_THREAD(t, run1);
  startClock();
  for(int i = 0; i<TEST1_N_LOCKS; i++)
  {
    BENCHMARK_INIT_LOCK(locks1[i]);
  }
  BENCHMARK_RUN_THREAD(t, run1);
  BENCHMARK_JOIN_THREAD(t);
  stopClock(1);
  return 1; 
}

FUNC_RET_TYPE run1(FUNC_ARG_TYPE)
{
  for(int i = 0; i<TEST1_N_LOCKS; i++)
  {
    BENCHMARK_LOCK(locks1[i]);
  }
  for(int i = 1; i<=TEST1_N_LOCKS; i++)
  {
    BENCHMARK_UNLOCK(locks1[TEST1_N_LOCKS-i]);
  }
#ifdef FUNC_RET_VAL
  return FUNC_RET_VAL;
#endif
}

int test2()
{
  BENCHMARK_THREAD_TYPE t; //Single thread
  BENCHMARK_INIT_THREAD(t, run2);
  startClock();
  BENCHMARK_INIT_LOCK(lock2);
  BENCHMARK_RUN_THREAD(t, run2);
  BENCHMARK_JOIN_THREAD(t);
  stopClock(2);
  return 1; 
}

FUNC_RET_TYPE run2(FUNC_ARG_TYPE)
{
  for(int i = 0; i<TEST2_N_RUNS; i++)
  {
    BENCHMARK_LOCK(lock2);
    BENCHMARK_UNLOCK(lock2);
  }
#ifdef FUNC_RET_VAL
  return FUNC_RET_VAL;
#endif
}

int test3()
{
  int attempt;
  testFinished = 0;
  BENCHMARK_INIT_LOCK(lock3A);
  BENCHMARK_INIT_LOCK(lock3B);
  BENCHMARK_THREAD_TYPE t; //Spawn thread
  BENCHMARK_INIT_THREAD(t, run3A);
  BENCHMARK_RUN_THREAD(t, run3A);
  for(attempt = 0; attempt < TEST3_MAX_ATTEMPTS; attempt++)
  {
    sleep(TEST3_TIMEOUT);
    if(testFinished)
      break;
    cout << "\tBenchmark test 3 is waiting for result..." << endl;
  }
  if(attempt == TEST3_MAX_ATTEMPTS)
  {
    cout << "FAILED: Protocol deadlocked in Benchmark Test 3!" << endl;
    return 0;
  }
  BENCHMARK_JOIN_THREAD(t);
  cout << "PASSED: Protocol did not deadlock in Benchmark Test 3." << endl;
  return 1;
}

FUNC_RET_TYPE run3A(FUNC_ARG_TYPE)
{
  BENCHMARK_THREAD_TYPE tSlave;
  BENCHMARK_LOCK(lock3A);
  BENCHMARK_INIT_THREAD(tSlave, run3B);
  BENCHMARK_RUN_THREAD(tSlave, run3B);
  sleep(TEST3_TIMEOUT); //Working! I.e. give slave time to lock lock3B
  BENCHMARK_LOCK(lock3B);
  BENCHMARK_UNLOCK(lock3B);
  BENCHMARK_UNLOCK(lock3A);
  BENCHMARK_JOIN_THREAD(tSlave);
  testFinished = 1;
#ifdef FUNC_RET_VAL
  return FUNC_RET_VAL;
#endif
}

FUNC_RET_TYPE run3B(FUNC_ARG_TYPE)
{
  BENCHMARK_LOCK(lock3B);
  BENCHMARK_LOCK(lock3A);
  BENCHMARK_UNLOCK(lock3A);
  BENCHMARK_UNLOCK(lock3B);
#ifdef FUNC_RET_VAL
  return FUNC_RET_VAL;
#endif
}

int test4()
{
  int attempt;
  testFinished = 0;
  BENCHMARK_INIT_LOCK(lock4A);
  BENCHMARK_INIT_LOCK(lock4B);
  BENCHMARK_THREAD_TYPE t; //Spawn thread
  BENCHMARK_INIT_THREAD(t, run4A);
  BENCHMARK_RUN_THREAD(t, run4A);
  for(attempt = 0; attempt < TEST4_MAX_ATTEMPTS; attempt++)
  {
    sleep(TEST4_TIMEOUT);
    if(testFinished)
      break;
    cout << "\tBenchmark test 4 is waiting for result..." << endl;
  }
  if(attempt == TEST4_MAX_ATTEMPTS)
  {
    cout << "FAILED: Protocol deadlocked or suffers from starvation in Benchmark Test 4!" << endl;
    return 0;
  }
  BENCHMARK_JOIN_THREAD(t);
  cout << "PASSED: Protocol did not deadlock or starve in Benchmark Test 4." << endl;
  return 1;
}

FUNC_RET_TYPE run4A(FUNC_ARG_TYPE)
{
  BENCHMARK_THREAD_TYPE tSlave;
  BENCHMARK_THREAD_TYPE tPartner;
  turn = 1;
  BENCHMARK_LOCK(lock4A);
  BENCHMARK_INIT_THREAD(tPartner, run4B);
  BENCHMARK_RUN_THREAD(tPartner, run4B);
  BENCHMARK_INIT_THREAD(tSlave, run4C);
  BENCHMARK_RUN_THREAD(tSlave, run4C);
  while(!testFinished)
  {
    if(turn == 1)
    {
      sleep(TEST4_TIMEOUT); //To give 4B enough time
                            //to reacquire lock4B
      BENCHMARK_UNLOCK(lock4A);
      turn = 2;
      BENCHMARK_LOCK(lock4A);
    }
  }
  BENCHMARK_UNLOCK(lock4A);
  BENCHMARK_JOIN_THREAD(tSlave);
  BENCHMARK_JOIN_THREAD(tPartner);
#ifdef FUNC_RET_VAL
  return FUNC_RET_VAL;
#endif
}

FUNC_RET_TYPE run4B(FUNC_ARG_TYPE)
{
  BENCHMARK_LOCK(lock4B);
  while(!testFinished)
  {
    if(turn == 2)
    {
      sleep(TEST4_TIMEOUT);
      BENCHMARK_UNLOCK(lock4B);
      turn = 1;
      BENCHMARK_LOCK(lock4B);
    }
  }
  BENCHMARK_UNLOCK(lock4B);
#ifdef FUNC_RET_VAL
  return FUNC_RET_VAL;
#endif
}

FUNC_RET_TYPE run4C(FUNC_ARG_TYPE)
{
  BENCHMARK_LOCK(lock4A);
  BENCHMARK_LOCK(lock4B);
  // DO WORK IN HERE
  BENCHMARK_UNLOCK(lock4B);
  BENCHMARK_UNLOCK(lock4A);
  testFinished = 1;
#ifdef FUNC_RET_VAL
  return FUNC_RET_VAL;
#endif
}

void startClock()
{
  struct timespec tempTime;
  clock_gettime(CLOCK_REALTIME, &tempTime); //Take start time
  startTime = tempTime.tv_nsec;
}

void stopClock(int testN)
{
  struct timespec tempTime;
  long endTime, resolution;
  double milliseconds;
  clock_gettime(CLOCK_REALTIME, &tempTime); //Take end time
  endTime = tempTime.tv_nsec;
  clock_getres(CLOCK_REALTIME, &tempTime); //and resolution
  resolution = tempTime.tv_nsec;
  milliseconds = (endTime-startTime)/1000000.0;
  cout << "Benchmark Test " << testN << " finishes with: " << (endTime-startTime) << "ns "
          "[Resolution: " << resolution << "ns] (= " << milliseconds << "ms)" << endl;
}

