
#include <cstdio>
#include <cstdlib>
#include <ctime>
#include <iostream>
using namespace std;

#include "benchmark.h"

#define LOCK_MAX 10      //How many lock/data pieces at most
#define THREAD_MAX 15    //How many threads at most
#define COLLATZ_MAX 1000 //Maximum of collatz seed
#define DATA_MAX 180     //According longest progression possible of collatz seed

#define VERBOSE 1        //Show messages on stdout

#ifdef GLOBAL_ATTR
GLOBAL_ATTR
#endif

typedef struct{
    BENCHMARK_LOCK_TYPE    lock;
    unsigned int data;
}DATA_STRUCT;

DATA_STRUCT *sec_data; //Array of locks and data
unsigned int len_data; //Length of array

unsigned int seed = 0;
unsigned int lock_max = LOCK_MAX;
unsigned int thread_max = THREAD_MAX;

long long startTime = 0LL;

FUNC_RET_TYPE run_thread(FUNC_ARG_TYPE); //Prototype for thread function
int collatz(int arg); //Next step in collatz conjecture

void startClock();
void stopClock();

int processArg(const char *arg, const char *val)
{
    if(arg[0]!='-' || arg[2]!='\0')
        return 0;
    switch(arg[1])
    {
        case 's': 
            return (sscanf(val, "%u", &seed) && seed > 0);
        case 'l':
            return (sscanf(val, "%u", &lock_max) && lock_max > 0);
        case 't':
            return (sscanf(val, "%u", &thread_max) && thread_max > 0); 
        default: return 0;
    }

}

int main(int argc, char *argv[])
{
    // * * * Variables * * *
    unsigned int i = 0;
    unsigned int nThreads = 0;
    BENCHMARK_THREAD_TYPE *threads;

#ifdef GLOBAL_INIT
GLOBAL_INIT()
#endif
    
    // * * * Read Arguments and place seed * * *//
    for(i = 2; (int)(i)<=argc; i+= 2)
    {
        if(argc%2 == 0 || processArg(argv[i-1], argv[i]) == 0)
        {
            cout << "Invalid argument given." << endl <<
            "Usage: ./benchmark [-s SEED>0 | -t THREADS>0 | -l LOCKS>0]" << endl;
            return 0;
        } 
    }
    if(seed == 0)
        seed = (unsigned int)(time(NULL));
    srand(seed);
    
    // * * * Create and init data * * *
    len_data = (unsigned int)(rand()%lock_max + 1); //1 to MAX locks
    sec_data = (DATA_STRUCT *)(new DATA_STRUCT[len_data]);
    for(i=0; i<len_data; i++)
    {
        BENCHMARK_INIT_LOCK(sec_data[i].lock);
        sec_data[i].data = 0;
    }

    // * * * Start a bunch of threads * * *
    nThreads = (unsigned int)(rand()%thread_max + 1); //1 to MAX threads
    threads = (BENCHMARK_THREAD_TYPE *)(new BENCHMARK_THREAD_TYPE[nThreads]);
    for(i=0; i<nThreads; i++)
    {
        BENCHMARK_INIT_THREAD(threads[i], run_thread);
    }
    startClock();
    for(i=0; i<nThreads; i++)
    {
        BENCHMARK_RUN_THREAD(threads[i], run_thread);
    }
    for(i=0; i<nThreads; i++)
    {
        BENCHMARK_JOIN_THREAD(threads[i]);
    }
    if(VERBOSE)
    {
        stopClock();
        cout << "Data:" << endl;
        for(i=0; i<len_data; i++)
            cout << sec_data[i].data << " ";
        cout << endl;
    }
    return 0;
}

FUNC_RET_TYPE run_thread(FUNC_ARG_TYPE)
{
    DATA_STRUCT *stack[DATA_MAX];
    unsigned int stackTop = 0;    //Index of the next free frame

    int i = 0;                    //Fill up an array of random numbers
    int nextRand = 0;             //for semi-randomness within 
    int semiRand[DATA_MAX];       //transactional memory
    for(i=0; i<DATA_MAX; i++)
        semiRand[i] = rand()%len_data;


    int c = (int)(rand()%COLLATZ_MAX + 1);
    bool inc;

#ifdef BENCHMARK_BEGIN_TRANS
    BENCHMARK_BEGIN_TRANS;
#endif
    while(c != 1)
    {
        inc = (c%3 != 0);
        if(inc)
        {
            int target = semiRand[nextRand++];
            stack[stackTop++] = &(sec_data[target]);
            BENCHMARK_LOCK(sec_data[target].lock);
            (sec_data[target].data) += 2;
        }
        else if(stackTop != 0)
        {
            (stack[--stackTop]->data)--;
            BENCHMARK_UNLOCK(stack[stackTop]->lock);
        }
        c = collatz(c);
    }
    while(stackTop > 0)
    {
        (stack[--stackTop]->data)--;
        BENCHMARK_UNLOCK(stack[stackTop]->lock);
    }
#ifdef BENCHMARK_END_TRANS
    BENCHMARK_END_TRANS;
#endif
#ifdef FUNC_RET_VAL
    return FUNC_RET_VAL;
#endif
}

int collatz(int arg)
{
    bool odd = (arg%2 == 1);
    if(arg <= 1)
        return 1;
    if(odd)
        return (3*arg+1);
    return (arg/2);
}

void startClock()
{
    struct timespec tempTime;
    clock_gettime(CLOCK_REALTIME, &tempTime); //Take start time
    startTime = (long long)(tempTime.tv_nsec);
}

void stopClock()
{
    struct timespec tempTime;
    long long endTime;
    long resolution;
    double milliseconds;
    clock_gettime(CLOCK_REALTIME, &tempTime); //Take end time
    endTime = tempTime.tv_nsec;
    clock_getres(CLOCK_REALTIME, &tempTime); //and resolution
    resolution = tempTime.tv_nsec;
    milliseconds = (endTime-startTime)/1000000.0;
    cout << "Tyche Benchmark Test finishes with: " << (endTime-startTime) << "ns "
            "[Resolution: " << resolution << "ns] (= " << milliseconds << "ms)" << endl;
}
   

 
