/**
 * Dynamic shared library which when loaded
 * replaces all pthread mutecies calls to
 * inherit locks. Please check the readme
 * file in this directory for further details.
 *
 * Uses a generic map as defined in map.h 
 *
 * @author: W. David Frohlingsdorf
 * @date: 2016
 * @purpose: Dissertation, University of Glasgow
 * */



/// * * * Thanks to * * *
/// -> stackoverflow.com/questions/3707358/get-all-the-thread-id-created-with-pthread-created-within-an-process

/// * * * Includes * * *

#define _GNU_SOURCE

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <bits/pthreadtypes.h>
#include <dlfcn.h>

#include "map.h"
#include "inheritance_lock_api.h"

/// * * * Definitions and Macros * * *

#define VERBOSE 2     //0=Off, 1=On, 2=Only Warnings and Errors
#define USE_PTHREAD 0 //Use pthread mutex rather than inheritance locks
#define GLIBC_VERSION "GLIBC_2.3.2"
#define COND

/// * * * Shared variables * * *

extern pthread_mutex_t graphLock;

/// * * * Redirected functions * * *

#undef pthread_mutex_lock

int pthread_mutex_lock(pthread_mutex_t *mutex)
{
  void *key = (void *)(mutex);
  INHERIT_LOCK lock = NULL;

  static int (*real_lock)(pthread_mutex_t *) = NULL;
  if(!real_lock)
    real_lock = dlsym(RTLD_NEXT, "pthread_mutex_lock");

  if(USE_PTHREAD || key == &graphLock)
    return real_lock(mutex);

  if(VERBOSE == 1)
    printf("Caught pthread_mutex_lock()\n");

  lock = map_get(key);
  if(lock == NULL)
  {
    lock = create_rec_lock(NULL);
    map_add(key, lock);
  }
  acquire(lock);
  return 0;
}

#undef pthread_mutex_unlock

int pthread_mutex_unlock(pthread_mutex_t *mutex)
{
  void *key = (void *)(mutex);
  INHERIT_LOCK lock = NULL;
  
  static int (*real_unlock)(pthread_mutex_t *) = NULL;
  if(!real_unlock)
    real_unlock = dlsym(RTLD_NEXT, "pthread_mutex_unlock");

  if(USE_PTHREAD || key == &graphLock)
    return real_unlock(mutex);
  
  if(VERBOSE == 1)
    printf("Caught pthread_mutex_unlock()\n");

  lock = map_get(key);
  release(lock);
  return 0;
}

#undef pthread_mutex_trylock

int pthread_mutex_trylock(pthread_mutex_t *mutex)
{
  static int (*real_trylock)(pthread_mutex_t *) = NULL;
  if(!real_trylock)
    real_trylock = dlsym(RTLD_NEXT, "pthread_mutex_trylock");

  if(USE_PTHREAD)
    return real_trylock(mutex);
  printf("TRYLOCK!\n");
  exit(0);
}

#undef pthread_mutex_destroy

int pthread_mutex_destroy(pthread_mutex_t *mutex)
{
  void *key = (void *)(mutex);
  INHERIT_LOCK lock = map_del(key);
  int r = 0;

  static int (*real_destroy)(pthread_mutex_t *) = NULL;
  if(!real_destroy)
    real_destroy = dlsym(RTLD_NEXT, "pthread_mutex_destroy");
  r = real_destroy(mutex);
  if(VERBOSE == 1)
    printf("Caught pthread_mutex_destroy()\n");
  if(lock != NULL)  //Only destroy if it has been initialised
    destroy_lock(&lock);
  return r;
}


/**********************CONDITION VARIABLE************************/
#ifdef COND
/****************************************************************/

#undef pthread_cond_timedwait

int pthread_cond_timedwait(pthread_cond_t *cond, pthread_mutex_t *mutex, const struct timespec *abstime)
{
  void *key = (void *)(mutex);
  char *err;
  static int (*real_timedwait)(pthread_cond_t *, pthread_mutex_t *, const struct timespec *) = NULL;
  if(!real_timedwait)
    real_timedwait = dlvsym(RTLD_NEXT, "pthread_cond_timedwait", GLIBC_VERSION);
  if(VERBOSE && (!real_timedwait || (err = dlerror())))
    printf("Fatal error: dlvsym() could not be executed correctly!\n"
           "\tError message: %s\n"
           "\tPlease check that you provided the correct version!", err); 


  if(VERBOSE && key != &graphLock)
  {
    struct timespec spec;
    struct timespec diff;
    clock_gettime(CLOCK_REALTIME, &spec);
    diff.tv_sec = abstime->tv_sec - spec.tv_sec;
    diff.tv_nsec = abstime->tv_nsec - spec.tv_nsec;
    printf("*****WARNING: pthread_cond_timedwait() has been called, timeout in: %ld[s],%ld[ns]*****\n", 
            (long)(diff.tv_sec), diff.tv_nsec);
  }
  printf("timedwait\n");
  return real_timedwait(cond, mutex, abstime);
}

#undef pthread_cond_wait

int pthread_cond_wait(pthread_cond_t *cond, pthread_mutex_t *mutex)
{
  void *key = (void *)(mutex);
  char *err;
  static int (*real_wait)(pthread_cond_t *, pthread_mutex_t *) = NULL;
  if(!real_wait)
    real_wait = dlvsym(RTLD_NEXT, "pthread_cond_wait", GLIBC_VERSION);
  if(VERBOSE && (!real_wait || (err = dlerror())))
    printf("Fatal error: dlvsym() could not be executed correctly!\n"
           "\tError message: %s\n"
           "\tPlease check that you provided the correct version!", err); 
  if(VERBOSE && key != &graphLock)
    printf("*****WARNING: pthread_cond_wait() has been called.*****\n");
  return real_wait(cond, mutex);
}

#undef pthread_cond_signal

int pthread_cond_signal(pthread_cond_t *cond)
{
  static int (*real_signal)(pthread_cond_t *) = NULL;
  char *err;
  if(!real_signal)
    real_signal = dlvsym(RTLD_NEXT, "pthread_cond_signal", GLIBC_VERSION);
  if(VERBOSE && (!real_signal || (err = dlerror())))
    printf("Fatal error: dlvsym() could not executed correctly!\n"
           "\tError message: %s\n"
           "\tPlease check that you provided the correct version!", err); 
  return real_signal(cond);
}

#endif
