/**
 * A generic map implemented with simple
 * single linked lists split by a primitive
 * hash function. 
 *
 * @author: W. David Frohlingsdorf
 * @date: 2016
 * @purpose: Dissertation, University of Glasgow
 * */

/// * * * Includes * * *

#include <stdlib.h>
#include <stdio.h>
#include "map.h"

/// * * * Defs and macros * * *

#define HASH_BOUND  262144 //2^18
#define VERBOSE     0

/// * * * Structures * * *

struct mapFrame{
  void *key;
  void *value;
  struct mapFrame *next;
};

/// * * * Typedefs * * *

typedef struct mapFrame frame;

/// * * * Static variables * * *

static frame *root[HASH_BOUND] = {NULL};
static unsigned long counter = 0;

/// * * * Local Functions * * *

//Recursive function to get the value point of a key
//@param Current frame
//@param Pointer to key
//@return Pointer to value or NULL if not in map
static void *get(frame *cur, void *key)
{
  if(cur == NULL)
    return NULL;      //Base case 1
  if(cur->key == key)
    return cur->value; //Base case 2
  return get(cur->next, key); //Inductive case
}


//Deletes an entry from the map and returns the value
//@param Previous frame (NULL if first)
//@param Current frame
//@param Pointer to key
//@return Pointer to value
static void *del(frame *pre, frame *cur, void *key)
{
  if(cur == NULL)
    return NULL;
  if(cur->key == key)
  {
    void *value;
    if(pre == NULL)
      root[(unsigned int)(key)%HASH_BOUND] = cur->next;
    else
      pre->next = cur->next;
    value = cur->value;
    free(cur);
    return value;
  }
  counter--;
  if(VERBOSE)
    printf("\tMap: Deleted lock now at: %lu\n", counter);
  return del(cur, cur->next, key);
}

/// * * * Header Functions * * *

void *map_get(void *key)
{
  return get(root[(unsigned int)(key)%HASH_BOUND], key);
}

void map_add(void *key, void *value)
{
  frame *newFrame = NULL;
  if(key == NULL || value == NULL)
    return;
  newFrame = (frame *)(malloc(sizeof(frame)));
  if(newFrame == NULL)
  {
    fprintf(stderr, "There has been a failed memory allocation! Aborting...\n");
    exit(EXIT_FAILURE);
  }
  newFrame->value = value;
  newFrame->key   = key;
  newFrame->next  = root[(unsigned int)(key)%HASH_BOUND];
  root[(unsigned int)(key)%HASH_BOUND] = newFrame;
  counter++;
  if(VERBOSE)
    printf("\tMap: Added lock now at: %lu\n", counter);
}

void *map_del(void *key)
{
  return del(NULL, root[(unsigned int)(key)%HASH_BOUND], key);
}


