/**
 * Inheritance Lock Protocol Source file
 *
 * @author: W. David Frohlingsdorf
 * @date: 2016
 * @purpose: Dissertation, University of Glasgow
 * */

#define _GNU_SOURCE //To allow MUTEX ERRORCHECK INITIALIZER

/// * * * * Standard Library Includes * * * *
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>

/// * * * * Inheritance Library Includes * * * *
#include "inheritance_lock_api.h"

/// * * * * Definitions and macros * * * *
#define STATIC_STACK_SIZE       32  //Size of the static stack of locks held by a thread
#define VERBOSE                 1   //Show messages on stderr if there was a problem in a
                                    //standard function
#define LOG_LEVEL               1   //0 No logging
                                    //1 Log only error messages
                                    //2 Log only inheritance locking
                                    //3 Log all locking
#define LOG_FILE "/users/level4/2079884f/locking.log"
#define LOG_VERBOSE             2   //0 Do not show log message on stdout
                                    //1 Show log message on stdout
                                    //2 Show log message on stdout and do not write to file

/// * * * * Structures * * * *

//Node for a lock in the wait-for graph
struct lockNode{
    struct threadNode *heldBy;      //Parent node, currently held by this thread
    struct threadNode *waitQueue;   //Head of the queue of waiting threads
    struct threadNode *tailQueue;   //Tail of the wait queue
    int recursive;                  //Boolean recursive lock
};

//Node for a thread in the wait-for graph
struct threadNode{
    struct lockNode *waitsFor;                              //Parent node, the lock this thread is waiting for
    struct lockNode *locksStaticStack[STATIC_STACK_SIZE];   //A stack of currently acquired locks of this thread
    unsigned int stackTop;                                  //Identifier holding the index of the entry one higher 
                                                            //than the last acquired lock (!)
    struct stackNode *locksDynamicStack;                    //If static stack is full (stackTop == STATIC_STACK_SIZE) use dynamic
    struct threadNode *waitQueueNext;                       //The next thread in the wait queue for the lock
    pthread_cond_t alarm;                                   //Conditional variable to wake up suspended thread
};

//A node for the dynamic stack used in the threadNode structure
struct stackNode{
    struct stackNode *prev; //Previous node in the stack. If NULL use static stack
    struct lockNode *lock;  //Lock assigned to this node
};


/// * * * * Typedefs * * * *

typedef struct lockNode LOCK;
typedef struct threadNode THRD;
typedef struct stackNode STCK;


/// * * * * Global Variables * * * *

static int volatile init = 0;                                   //Boolean to indicate if the system has been initialised
pthread_mutex_t graphLock = PTHREAD_ERRORCHECK_MUTEX_INITIALIZER_NP; 
                                                                //Mutex to allow mutual exclusive access to the wait-for graph
                                                                //Not static to allow shared library to access variable
static pthread_key_t nodeKey;                                   //Thread specific key to find the according threadNode
static pthread_once_t initKey = PTHREAD_ONCE_INIT;              //Once key to initialise nodeKey

/// * * * * Prototypes * * * *

static void lockGraph();
static void unlockGraph();
static int release_intern(LOCK *l, THRD *t);
static INHERIT_LOCK build_lock(int recursive, INHERIT_LOCK *pLock);
static THRD *getNode();
static void threadDtor(void *node);
static void initaliseNodeKey();
static LOCK *pop(THRD *t);
static void push(THRD *t, LOCK *l);
static LOCK *peek(THRD *t);
static int stackContains(THRD *t, LOCK *l);
static THRD *getRoot(LOCK *l);
static void enqueue(LOCK *l, THRD *t);
static THRD *dequeue(LOCK *l);

static void abortWithExcept(const char *description, int error);
static void outOfMemExcept(const char *functionName, int line);

/// * * * * Implementations of static functions * * * *

//Logger
//@param String to be written to log file
void logger(const char *message)
{
    if(LOG_VERBOSE >= 1)
        printf("%s", message);
    if(LOG_VERBOSE < 2)
    {
        FILE *fp = fopen(LOG_FILE, "a");
        if(fp == NULL)
            fprintf(stderr, "Unable to open log file at: %s\n", LOG_FILE);
        else
        {
            fprintf(fp, "%s", message);
            fclose(fp);
        }
    }
}

//Locks the wait-for graph
void lockGraph()
{
    int e;
    e = pthread_mutex_lock(&graphLock);
    if(e)
        abortWithExcept("Could not lock the Graph! pthread_mutex_lock() returned with", e);
}

//Unlocks the wait-for graph
void unlockGraph()
{
    int e;
    e = pthread_mutex_unlock(&graphLock);
    if(e)
        abortWithExcept("Could not unlock the Graph! pthread_mutex_unlock() returned with", e);
}

//Releases a lock from the thread.
//Other than the API release, this internal release function requires the 
//thread structure to be handed over in the second parameter. This is necessary
//as release_intern() might be called from the thread specific memory desctructor.
//@param The according lock
//@param The according thread
//@return -1 if an error occured, 0 otherwise
int release_intern(LOCK *lock, THRD *t)
{
    THRD *next;
    int e;
    if(peek(t) != lock)     //We can only release the lock from the top of the stack
    {
        if(VERBOSE)
            fprintf(stderr, "Tried to release a lock which is not on top of the stack!\n");
        if(LOG_LEVEL >= 1)
            logger("Tried to release a lock which is not on top of the stack!\n");
        return -1;
    }
    
    
    lockGraph();
    pop(t);
    if(t != lock->heldBy)   //Thread inherited this lock
    {
        if(LOG_LEVEL >= 2)
            logger("Released lock inheritently.\n");
        unlockGraph();
        return 0;
    }
    next = dequeue(lock);
    lock->heldBy = next;
    if(next!=NULL)          //If another thread is waiting for the lock, hand it over and wake up the thread
    {
        next->waitsFor = NULL;
        push(next, lock);
        e = pthread_cond_signal(&(next->alarm));
        if(e)
            abortWithExcept("An error occured while signaling condition variable!" 
                            "pthread_cond_signal() returned with", e);
        if(LOG_LEVEL >= 3)
            logger("Transfered lock to next thread in queue.\n");
    }
    else if(LOG_LEVEL >= 3)
        logger("Released lock normally.\n");
    unlockGraph();
    return 0;
}

//Allocates memory and initializes an inheritance lock
//@param Boolean if the lock should be recursive
//@param Optional pointer which will be set to point to the newly created lock
INHERIT_LOCK build_lock(int recursive, INHERIT_LOCK *l)
{
    LOCK *lock = (LOCK *)(malloc(sizeof(LOCK)));
    if(lock == NULL)
        outOfMemExcept("INHERIT_LOCK build_lock(int, INHERIT_LOCK *)", __LINE__);
    lock->heldBy = NULL;
    lock->waitQueue = NULL;
    lock->tailQueue = NULL;
    lock->recursive = recursive;
    
    if(l != NULL)
        (*l) = (INHERIT_LOCK)(lock);
    return (INHERIT_LOCK)(lock);
}

//Function which returns the thread specific wait-for graph node of the calling thread
THRD *getNode()
{
    int e;
    
    pthread_once(&initKey, initaliseNodeKey);    //The first thread should initalize the node key
    while(!init)                                 //since this might happen concurrently we 
        ;                                        //ensure the race condition with a busy wait
    
    void *n = pthread_getspecific(nodeKey);
    if(n == NULL)                                //First time this thread calls this function
    {                                            //Initialize the node
        THRD *t = (THRD *)(malloc(sizeof(THRD)));
        if(t == NULL)
            outOfMemExcept("THRD *getNode()", __LINE__);
        t->waitsFor = NULL;
        t->stackTop = 0;
        t->locksDynamicStack = NULL;
        t->waitQueueNext = NULL;
        e = pthread_cond_init(&(t->alarm), NULL);
        if(e)
            abortWithExcept("Could not initialise condition variable! pthread_cond_init() returned with", e);
        n = (void *)(t);
        e = pthread_setspecific(nodeKey, n);
        if(e)
            abortWithExcept("Could not set thread specific object! pthread_setspecific() returned with", e);
    }
    return (THRD *)(n);
}

//Destructor of thread specific wait-for graph node
//This function should be automatically called at the end of a pthread's life
//If there are still locks assigned to the thread we will try to release them first
//@param The value associated with the key. Setspecific is already set to NULL
void threadDtor(void *thread)
{
    int e;
    if(!init)
        abortWithExcept("Thread destructor called before key has been initialised", 0);
    THRD *t = (THRD *)(thread);
    if(t != NULL)
    {
        LOCK *l;
        while((l = peek(t)) != NULL)
        {
            if(release_intern(l, t))
                abortWithExcept("Thread destructor unable to release a held lock! Fix your code!", 0);
        }
        e = pthread_cond_destroy(&(t->alarm));
        if(e)
            abortWithExcept("Unable to destroy condition variable! pthread_cond_destroy() returned with", e);
        free(t);
    }
}

//Called only once. Creates thread specific data accessible through nodeKey
void initaliseNodeKey()
{
    int e = pthread_key_create(&nodeKey, threadDtor);
    if(e)
        abortWithExcept("Could not create pthread key! pthread_key_create() returned with", e);
    init = 1;
}

//Pops the top of the lock-stack of currently held locks of the thread
//The function automatically deals with static and dynamic stack
//@param The wait-for node of the thread
LOCK *pop(THRD *t)
{
    if(t->stackTop == 0)                   //The stack is empty
        return NULL;
    if(t->locksDynamicStack != NULL)       //Dynamic stack is used
    {
        LOCK *l = NULL;
        STCK *s = t->locksDynamicStack;
        t->locksDynamicStack = s->prev;
        l = s->lock;
        free(s);
        return l;
    }
    return t->locksStaticStack[--(t->stackTop)];
}

//Push a lock on top of the thread specific wait-for node
//Automatically deals with static and dynamic stack
//@param The wait-for node of the thread
//@param The wait-for node of the according lock
void push(THRD *t, LOCK *l)
{
    if(t->stackTop == STATIC_STACK_SIZE)
    {
        STCK *frame = (STCK *)(malloc(sizeof(STCK)));
        if(frame == NULL)
            outOfMemExcept("void push(THRD *, LOCK *)", __LINE__);
        frame->lock = l;
        frame->prev = t->locksDynamicStack;
        t->locksDynamicStack = frame;
    }
    else
        t->locksStaticStack[t->stackTop++] = l;
}

//Like pop, but leaves the top of the stack on
//@param The wait-for node of the thread
LOCK *peek(THRD *t)
{
    if(t->stackTop == 0)
        return NULL;
    if(t->locksDynamicStack != NULL)
        return (t->locksDynamicStack->lock);
    return t->locksStaticStack[(t->stackTop)-1];
}

//Checks if the stack contains the lock node
//Needed for non-recursive locks
//@param The wait-for node of the thread
//@param The node of the lock
//@return 1 if stack contains the node, 0 otherwise
int stackContains(THRD *t, LOCK *l)
{
    STCK *frame;
    int cursor;
    for(frame = t->locksDynamicStack; frame != NULL; frame = frame->prev)
    {
        if(frame->lock == l)
            return 1;
    }
    for(cursor = (t->stackTop)-1; cursor >= 0; cursor--)
    {
        if(t->locksStaticStack[cursor] == l)
            return 1;
    }
    return 0;
}

//Find the root of the tree which the lock is assigned to
//That is, find the thread currently holding (directly or indirectly
//through inheritance) the lock
//@param The node of the lock
//@return NULL if not held by any thread, the wait-for node of the according thread otherwise
THRD *getRoot(LOCK *l)
{
    THRD *t = l->heldBy;
    if(t == NULL)
        return NULL;
    while(t->waitsFor != NULL)
    {
        l = t->waitsFor;     //If the thread is waiting for a lock
        t = l->heldBy;       //it must be held by a different thread!
    }
    return t;
}

//Enqueues a thread in the wait queue for a lock
//@param The node of the lock
//@param The node of the thread
void enqueue(LOCK *l, THRD *t)
{
    THRD *c = l->tailQueue;
    if(c == NULL)            //First one in the queue
        l->waitQueue = t;
    else
        c->waitQueueNext = t;
    l->tailQueue = t;
    t->waitQueueNext = NULL;
}

//Dequeues the head of the waiting queue for this lock
//@param The according lock
//@return The wait-for node of the thread, if none is waiting NULL
THRD *dequeue(LOCK *l)
{
    THRD *c = l->waitQueue;
    if(c == NULL)
        return NULL;
    l->waitQueue = c->waitQueueNext;
    if(l->waitQueue == NULL)
        l->tailQueue = NULL;
    c->waitQueueNext = NULL;
    return c;
}

//Throws an error message and aborts execution
//@param Error description
//@param Error code
void abortWithExcept(const char *desc, int err)
{
    fprintf(stderr, "Fatal error: %s %d\nAborting...", desc, err);
    exit(EXIT_FAILURE);
}

//Throws an out of memory error (if malloc failed)
//@param Function name where the error happend
//@param Line number from which function is called
void outOfMemExcept(const char *funcName, int line)
{
    fprintf(stderr, "Fatal error: Out of memory! In function %s [Line: %d]\nAborting...", funcName, line);
    exit(EXIT_FAILURE);
}

        

/// * * * * Implementations of API functions * * * *
/// A more detailed description of each funciton is in the inheritance_lock_api.h
/// header file


INHERIT_LOCK create_lock(INHERIT_LOCK *l)
{
    return build_lock(0, l);
}

INHERIT_LOCK create_rec_lock(INHERIT_LOCK *l)
{
    return build_lock(1, l);
}

int acquire(INHERIT_LOCK l)
{
    LOCK *lock = (LOCK *)(l);
    THRD *t = getNode();
    THRD *root;
    int e;
    if(l == NULL)
    {
        if(VERBOSE)
            fprintf(stderr, "Can not acquire a NULL pointer!\n");
        if(LOG_LEVEL >= 1)
            logger("Can not acquire a NULL pointer!\n");
        return -1;
    }
    if(lock->recursive == 0 && lock->heldBy != NULL && stackContains(t, lock))
    {
        if(VERBOSE)
            fprintf(stderr, "Non-recursive lock has already been acquired before!\n");
        if(LOG_LEVEL)
            logger("Non-recursive lock has already been acquired before!\n");
        return -1;          //Non-recursive lock has already been acquired before
    }
    
    lockGraph();
    root = getRoot(lock);   //Check which thread is currently holding the lock
    if(root == NULL)        //No one is holding this lock
    {
        push(t, lock);
        lock->heldBy = t;
        if(LOG_LEVEL >= 3)
            logger("Acquired lock normally.\n");
        unlockGraph();
        return 0;
    }
    if(root == t)           //This thread is already holding the lock
    {
        push(t, lock);
        if(LOG_LEVEL >= 2)
            logger("Acquired lock inheritently.\n");
        unlockGraph();
        return 0;
    }
    t->waitsFor = lock;     //A different thread is holding the lock
    enqueue(lock, t);
    if(LOG_LEVEL >= 3)
        logger("Waiting for lock to become available.\n");
    e = pthread_cond_wait(&(t->alarm), &graphLock);
    if(e)
        abortWithExcept("An error occured in pthread_cond_wait(). Returned with", e);
    unlockGraph();
    return 0;
}

int release(INHERIT_LOCK l)
{
    LOCK *lock = (LOCK *)(l);
    THRD *t = getNode();
    return release_intern(lock, t);
}

int destroy_lock(INHERIT_LOCK *l)
{
    if(l == NULL)
    {
        if(VERBOSE)
            fprintf(stderr, "Tried to destroy lock with NULL pointer!\n");
        if(LOG_LEVEL >= 1)
            logger("Tried to destroy lock with NULL pointer!\n");
        return -1;
    }
    
    LOCK *lock = (LOCK *)(*l);
    if(lock == NULL || lock->heldBy != NULL || lock->waitQueue != NULL)
    {
        if(VERBOSE)
            fprintf(stderr, "Tried to destroy lock which is still in use!\n");
        if(LOG_LEVEL >= 1)
            logger("Tried to destroy lock which is still in use!\n");
        return -1;          //Can only destroy a lock if no thread is holding it and is neither waiting for it
    }
    free(lock);
    (*l) = NULL;
    return 0;
}

